package ru.dbotthepony.kstarbound.lua.userdata

import com.google.gson.JsonNull
import com.google.gson.JsonObject
import org.classdump.luna.ByteString
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import org.classdump.luna.Userdata
import org.classdump.luna.impl.ImmutableTable
import org.classdump.luna.runtime.LuaFunction
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.actor.behavior.BehaviorDefinition
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.iterator
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.tableMapOf
import ru.dbotthepony.kstarbound.lua.toJson
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.world.entities.behavior.AbstractBehaviorNode
import ru.dbotthepony.kstarbound.world.entities.behavior.BehaviorTree
import ru.dbotthepony.kstarbound.world.entities.behavior.Blackboard
import java.util.concurrent.atomic.AtomicLong

class BehaviorState(val tree: BehaviorTree) : Userdata<BehaviorState>() {
	val blackboard get() = tree.blackboard
	val lua get() = blackboard.lua
	val functions = HashMap<String, LuaFunction<*, *, *, *, *>>()

	init {
		lua.attach(tree.scripts)

		for (name in tree.functions) {
			val get = lua.globals[name]

			if (get == null) {
				throw LuaRuntimeException("No such function for behavior: $name")
			} else if (get !is LuaFunction<*, *, *, *, *>) {
				throw LuaRuntimeException("Not a Lua function for behavior: $name")
			} else {
				this.functions[name] = get
			}
		}
	}

	fun run(delta: Double): AbstractBehaviorNode.Status {
		val ephemerals = blackboard.takeEphemerals()
		val status = tree.runAndReset(delta, this)
		blackboard.clearEphemerals(ephemerals)
		return status
	}

	override fun getMetatable(): Table {
		return Companion.metatable
	}

	override fun setMetatable(mt: Table?): Table {
		throw UnsupportedOperationException()
	}

	override fun getUserValue(): BehaviorState {
		return this
	}

	override fun setUserValue(value: BehaviorState?): BehaviorState? {
		throw UnsupportedOperationException()
	}

	companion object {
		// required for Lua code
		// in original engine, passes directly 32/64 bit pointer of *Node struct
		val NODE_GARBAGE_INDEX = AtomicLong()

		private fun __index(): Table {
			return metatable
		}

		private val metatable = ImmutableTable.Builder()
			.add("__index", luaFunction { _: Any?, index: Any -> returnBuffer.setTo(__index()[index]) })
			.add("run", luaFunction { self: BehaviorState, delta: Number ->
				returnBuffer.setTo(self.run(delta.toDouble()))
			})
			.add("clear", luaFunction { self: BehaviorState ->
				self.tree.reset()
			})
			.add("blackboard", luaFunction { self: BehaviorState ->
				returnBuffer.setTo(self.blackboard)
			})
			.build()

		fun provideBindings(lua: LuaEnvironment) {
			lua.globals["behavior"] = lua.tableMapOf(
				"behavior" to luaFunction { config: Any, parameters: Table, _: Any?, blackboard: Blackboard? ->
					val tree: BehaviorTree

					if (config is ByteString) {
						val base = Registries.behavior.getOrThrow(config.decode())

						if (!parameters.iterator().hasNext()) {
							tree = BehaviorTree(blackboard ?: Blackboard(lua), base.value)
						} else {
							tree = BehaviorTree(blackboard ?: Blackboard(lua), Starbound.gson.fromJsonFast(base.json.deepCopy().also {
								it as JsonObject
								it["parameters"] = mergeJson(it["parameters"] ?: JsonNull.INSTANCE, toJsonFromLua(parameters))
							}, BehaviorDefinition::class.java))
						}
					} else {
						val cast = (config as Table).toJson(true) as JsonObject

						if (parameters.iterator().hasNext())
							cast["parameters"] = mergeJson(cast["parameters"] ?: JsonNull.INSTANCE, toJsonFromLua(parameters))

						tree = BehaviorTree(blackboard ?: Blackboard(lua), Starbound.gson.fromJsonFast(cast, BehaviorDefinition::class.java))
					}

					returnBuffer.setTo(BehaviorState(tree))
				}
			)
		}
	}
}
