package ru.dbotthepony.kstarbound.world.entities

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import org.apache.logging.log4j.LogManager
import org.classdump.luna.ByteString
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import ru.dbotthepony.kommons.collect.ListenableMap
import ru.dbotthepony.kommons.collect.collect
import ru.dbotthepony.kommons.collect.map
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.io.IntValueCodec
import ru.dbotthepony.kommons.io.StreamCodec
import ru.dbotthepony.kommons.io.nullable
import ru.dbotthepony.kommons.io.readKOptional
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.io.writeKOptional
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kommons.util.value
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.collect.IdMap
import ru.dbotthepony.kstarbound.defs.DamageData
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.EphemeralStatusEffect
import ru.dbotthepony.kstarbound.defs.StatusEffectDefinition
import ru.dbotthepony.kstarbound.defs.actor.PersistentStatusEffect
import ru.dbotthepony.kstarbound.defs.actor.StatModifier
import ru.dbotthepony.kstarbound.defs.actor.StatModifierType
import ru.dbotthepony.kstarbound.defs.actor.StatusControllerConfig
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.LuaMessageHandlerComponent
import ru.dbotthepony.kstarbound.lua.LuaUpdateComponent
import ru.dbotthepony.kstarbound.lua.bindings.MovementControllerBindings
import ru.dbotthepony.kstarbound.lua.bindings.createConfigBinding
import ru.dbotthepony.kstarbound.lua.bindings.provideAnimatorBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideConfigBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideEntityBindings
import ru.dbotthepony.kstarbound.lua.bindings.provideWorldBindings
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.iterator
import ru.dbotthepony.kstarbound.lua.luaFunction
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.network.packets.DamageNotificationPacket
import ru.dbotthepony.kstarbound.network.syncher.NetworkedDynamicGroup
import ru.dbotthepony.kstarbound.network.syncher.NetworkedGroup
import ru.dbotthepony.kstarbound.network.syncher.NetworkedElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedMap
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedFloat
import ru.dbotthepony.kstarbound.network.syncher.networkedJsonObject
import ru.dbotthepony.kstarbound.network.syncher.networkedString
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.util.HistoryQueue
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.api.StatusEffectEntity
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.Collections
import java.util.stream.Collectors

// this is unnatural to have this class separated, but since it contains
// lots of internal state, it would be nice to have it encapsulated;
// despite this class being used exactly once - as part of ActorEntity.

// On other hand, original code separates StatCollection into subclass,
// and we won't do this.

// On side note, it appears game initially separated StatSet from StatCollection,
// which means early in development Starbound had true singleplayer, which eventually
// got removed in favor of listening server.
// Unfortunately for me, this means I have to untangle this piece of code
// and unify StatSet and StatCollection logic into one thing.

// StatSet original docs:
// Manages a collection of Stats and Resources.
//
// Stats are named floating point values of any base value, with an arbitrary
// number of "stat modifiers" attached to them.  Stat modifiers can be added
// and removed in groups, and they can either raise or lower stats by a
// constant value or a percentage of the stat value without any other
// percentage modifications applied.  The effective stat value is always the
// value with all mods applied.  If a modifier is created for a stat that does
// not exist, there will be an effective stat value for the modified stat, but
// NO base stat.  If the modifier is a base percentage modifier, it will have
// no effect because it is assumed that base stats that do not exist are zero.
//
// Resources are also named floating point values, but are in a different
// namespaced and are intended to be used as values that change regularly.
// They are always >= 0.0f, and optionally have a maximum value based on a
// given value or stat.  In addition to a max value, they can also have a
// "delta" value or stat, which automatically adds or removes that delta to the
// resource every second.
//
// If a resource has a maximum value, then rather than trying to keep the
// *value* of the resource constant, this class will instead attempt to keep
// the *percentage* of the resource constant across stat changes.  For example,
// if "health" is a stat with a max of 100, and the current health value is 50,
// and the max health stat is changed to 200 through any means, the health
// value will automatically update to 100.

// StatCollection original docs:
// Extension of StatSet that can easily be set up from config, and is network
// capable.
class StatusController(val entity: ActorEntity, val config: StatusControllerConfig) : NetworkedElement.Passthrough() {
	// status effects
	private val networkGroup = NetworkedGroup()
	private val statNetworkGroup = NetworkedGroup()

	override val parentElement: NetworkedElement
		get() = networkGroup

	init {
		networkGroup.add(statNetworkGroup)
	}

	val lua = LuaEnvironment()
	val luaUpdate = LuaUpdateComponent(lua)
	val luaMovement = MovementControllerBindings(entity.movement)
	val luaMessages = LuaMessageHandlerComponent(lua) { toString() }

	private val animator: EffectAnimator?
	private val animatorID: Int?

	fun init() {
		if (!entity.isRemote) {
			luaUpdate.stepCount = config.primaryScriptDelta
			lua.attach(config.primaryScriptSources)

			// provideStatusControllerBindings(this, lua) // provided through provideEntityBindings
			provideEntityBindings(entity, lua)
			luaMovement.init(lua)

			// TODO: Once we have brand new object-oriented Lua API, expose proper entity bindings here
			// TODO: Expose world bindings
			lua.init()
		}
	}

	fun applyMovementControls() {
		luaMovement.apply()
		uniqueStatusEffects.values.forEach { it.luaMovement.apply() }
	}

	// TODO: do we really need this? Original engine uses this for reducing CPU pressure,
	//  but if we just fix inefficiencies in code inside if block, this won't be required
	private val environmentalUpdateTimer = GameTimer(config.environmentStatusEffectUpdateTimer)

	private val recentHitsDealt = HistoryQueue<Pair<Int, DamageData>>(config.keepDamageNotificationSteps)
	private val recentDamageReceived = HistoryQueue<DamageNotification>(config.keepDamageNotificationSteps)
	private val recentDamageDealt = HistoryQueue<DamageNotification>(config.keepDamageNotificationSteps)

	fun notifyHitDealt(victimID: Int, damage: DamageData) {
		recentHitsDealt.add(victimID to damage)
	}

	fun notifyDamageDealt(damage: DamageNotification) {
		recentDamageDealt.add(damage)
	}

	fun recentHitsDealt(since: Long = 0L) = recentHitsDealt.query(since)
	fun recentDamageReceived(since: Long = 0L) = recentDamageReceived.query(since)
	fun recentDamageDealt(since: Long = 0L) = recentDamageDealt.query(since)

	fun experienceDamage(damage: DamageData): List<DamageNotification> {
		val results = lua.invokeGlobal("applyDamageRequest", lua.from(Starbound.gson.toJsonTree(damage)))

		if (results.isNotEmpty() && results[0] is Table) {
			val parsed = ArrayList<DamageNotification>()

			for ((_, v) in (results[0] as Table)) {
				try {
					parsed.add(Starbound.gson.fromJsonFast(toJsonFromLua(v), DamageNotification::class.java))
				} catch (err: Throwable) {
					LOGGER.error("Exception while parsing results returned by applyDamageRequest (primary scripts: ${config.primaryScriptSources})", err)
				}
			}

			for (notification in parsed) {
				recentDamageReceived.add(notification)
			}

			return parsed
		}

		return listOf()
	}

	// since experienceDamage is expected to be called externally (when someone damaged us)
	// inflicting self damage requires little more work (notifying everyone else that we received damage)
	fun inflictSelfDamage(damage: DamageData) {
		val notifications = experienceDamage(damage)

		for (notification in notifications) {
			// addDamageNotification instead of networkDamageNotification for original engine parity
			// (inflicting self-damage calls "damagedOther" functions on original engine, so as well must we)
			entity.world.addDamageNotification(DamageNotificationPacket(entity.entityID, notification))
		}
	}

	// used as place to store random data by Lua scripts, because global `storage` table wasn't enough, it seems,
	// to Chucklefish devs
	private val statusProperties = networkedJsonObject(config.statusProperties.deepCopy()).also { networkGroup.add(it) }

	fun setProperty(key: String, value: JsonElement) {
		val old = statusProperties.value[key]
		statusProperties.value[key] = value

		if (old != value) {
			statusProperties.bumpVersion()
		}
	}

	fun getProperty(key: String): JsonElement? {
		return statusProperties.value[key]
	}

	var parentDirectives by networkedString().also { networkGroup.add(it) }
		private set

	var primaryDirectives = ""

	var appliesEnvironmentStatusEffects = config.appliesEnvironmentStatusEffects
	var appliesWeatherStatusEffects = config.appliesWeatherStatusEffects
	var minimumLiquidStatusEffectPercentage = config.minimumLiquidStatusEffectPercentage

	private val uniqueEffectMetadata = NetworkedDynamicGroup(::UniqueEffectNetworkedValues, UniqueEffectNetworkedValues::networkGroup).also { networkGroup.add(it) }
	private val effectAnimators = NetworkedDynamicGroup(::EffectAnimator, { it }).also { networkGroup.add(it) }

	init {
		if (config.primaryAnimationConfig == null) {
			animator = null
			animatorID = null
		} else {
			animator = EffectAnimator(KOptional(config.primaryAnimationConfig.fullPath))
			animatorID = effectAnimators.add(animator)

			// FIXME: effect animator's animator is not set in stone because of legacy protocol
			//  god damn it
			//  But at least it shouldn't change in this context
			if (!entity.isRemote) {
				provideAnimatorBindings(animator.animator, lua)
			}
		}
	}

	fun handleMessage(message: String, isLocal: Boolean, arguments: JsonArray): JsonElement? {
		return luaMessages.handle(message, isLocal, arguments) ?: uniqueStatusEffects.values.firstNotNullOfOrNull { it.luaMessages.handle(message, isLocal, arguments) }
	}

	@JsonFactory
	data class SerializedData(
		val statusProperties: JsonObject,
		val resourceValues: ImmutableMap<String, Double>,
		val resourcesLocked: ImmutableMap<String, Boolean>,
		val persistentEffectCategories: ImmutableMap<String, ImmutableList<PersistentStatusEffect>>,
		val ephemeralEffects: ImmutableList<UniqueEffect>,
	) {
		@JsonFactory
		data class UniqueEffect(
			val effect: Registry.Ref<StatusEffectDefinition>,
			val duration: Double,
			val maxDuration: Double = duration,
		)
	}

	fun serialize(): SerializedData {
		return SerializedData(
			statusProperties = statusProperties.get(),
			resourceValues = resourcesInternal.entries.stream().map { it.key to it.value.value }.collect(ImmutableMap.toImmutableMap({ it.first }, { it.second })),
			resourcesLocked = resourcesInternal.entries.stream().map { it.key to it.value.isLocked }.collect(ImmutableMap.toImmutableMap({ it.first }, { it.second })),

			ephemeralEffects = uniqueStatusEffects
				.values
				.stream()
				.filter { !it.isPersistent }
				.map {
					SerializedData.UniqueEffect(
						effect = it.effect.ref,
						duration = it.metadata.duration,
						maxDuration = it.metadata.maxDuration,
					)
				}
				.collect(ImmutableList.toImmutableList()),

			persistentEffectCategories = persistentStatusEffects
				.entries
				.stream()
				.map {
					val result = ImmutableList.Builder<PersistentStatusEffect>()
					it.value.statModifiers.forEach { result.add(Either.left(it)) }
					it.value.uniqueEffects.forEach { result.add(Either.right(it.ref)) }
					it.key to result.build()
				}
				.collect(ImmutableMap.toImmutableMap({ it.first }, { it.second })),
		)
	}

	fun deserialize(data: SerializedData) {

	}

	// ----- Persistent status effects
	// These effects have no duration and are typically sourced to actor's equipment OR
	// the environment. Sometimes, they are also sources to actor's current action, such as sleeping on bed.
	private class PersistentEffectCategory(val statGroupID: Int) {
		val statModifiers = ArrayList<StatModifier>()
		val uniqueEffects = HashSet<Registry.Entry<StatusEffectDefinition>>()
	}

	// there are two magic keys used for this map: 'entities' and 'environment' for StatusEffectEntity
	// and environmentally applied persistent status effects, respectively
	private val persistentStatusEffects = HashMap<String, PersistentEffectCategory>()

	fun addPersistentEffect(category: String, effect: PersistentStatusEffect) {
		addPersistentEffects(category, listOf(effect))
	}

	fun addPersistentEffects(category: String, effects: Collection<PersistentStatusEffect>) {
		val cat = persistentStatusEffects.computeIfAbsent(category) {
			val index = statModifiers.nextFreeIndex()
			val value = PersistentEffectCategory(index)
			// avoid double-copy
			statModifiersNetworkMap[index] = value.statModifiers
			value
		}

		for (effect in effects) {
			effect.map({ cat.statModifiers.add(it) }, { if (it.isPresent) cat.uniqueEffects.add(it.entry!!) })
		}

		statModifiersNetworkMap.markIndexDirty(cat.statGroupID)
		updatePersistentUniqueEffects()
	}

	fun setPersistentEffects(category: String, effects: Collection<PersistentStatusEffect>) {
		var changes = false

		if (effects.isEmpty() && category in persistentStatusEffects) {
			statModifiersNetworkMap.remove(persistentStatusEffects.remove(category)!!.statGroupID)
			changes = true
		} else if (effects.isNotEmpty()) {
			val cat = persistentStatusEffects[category]

			if (cat != null) {
				cat.uniqueEffects.clear()
				cat.statModifiers.clear()
			}

			addPersistentEffects(category, effects)
			changes = true
		}

		if (changes)
			updatePersistentUniqueEffects()
	}

	fun removePersistentEffects(category: String) {
		setPersistentEffects(category, listOf())
	}

	fun removeAllPersistentEffects() {
		if (persistentStatusEffects.isNotEmpty()) {
			persistentStatusEffects.values.forEach {
				statModifiersNetworkMap.remove(it.statGroupID)
			}

			persistentStatusEffects.clear()
			updatePersistentUniqueEffects()
		}
	}

	private fun updatePersistentUniqueEffects() {
		val activePersistentEffects = HashSet<Registry.Entry<StatusEffectDefinition>>()

		persistentStatusEffects.values.forEach {
			activePersistentEffects.addAll(it.uniqueEffects)
		}

		for (effect in activePersistentEffects) {
			val existing = uniqueStatusEffects[effect.key]

			if (existing == null) {
				addUniqueEffect(effect, null, null)
			} else {
				existing.promoteToPersistent()
			}
		}

		// Again, here we are using "durationless" to mean "persistent"
		val keys = ArrayList(uniqueStatusEffects.keys)

		for (key in keys) {
			val meta = uniqueStatusEffects[key]!!

			if (meta.metadata.duration == 0.0 && meta.effect !in activePersistentEffects) {
				meta.remove()
			}
		}
	}

	fun getPersistentEffects(category: String): List<PersistentStatusEffect> {
		val cat = persistentStatusEffects[category] ?: return emptyList()
		val result = ArrayList<PersistentStatusEffect>()

		cat.statModifiers.forEach {
			result.add(Either.left(it))
		}

		cat.uniqueEffects.forEach {
			result.add(Either.right(it.ref))
		}

		return result
	}

	// ----- Unique status effects
	// These are the effect player see on their status bar (under health).
	// Difference between StatModifiers and these are that Unique status effect can have Lua logic attached,
	// and don't necessarily affect stats (but frequently come with stat modifiers, such as providing health regeneration
	// through providing positive 'delta' value for 'health' resource)
	private val uniqueStatusEffects = LinkedHashMap<String, UniqueEffectInstance>()

	val isStatusEffectsImmune: Boolean
		get() = statPositive("statusImmunity")

	private inner class UniqueEffectInstance(val effect: Registry.Entry<StatusEffectDefinition>, duration: Double?, source: Int?) {
		init {
			if (effect.key in uniqueStatusEffects) {
				throw IllegalStateException("Already has unique effect $effect")
			}

			uniqueStatusEffects[effect.key] = this
		}

		var parentDirectives: String = ""
		val modifierGroups = IntArrayList()

		val lua = LuaEnvironment()
		val luaMessages = LuaMessageHandlerComponent(lua) { "Unique effect" }
		val luaMovement = MovementControllerBindings(entity.movement)
		val luaUpdate = LuaUpdateComponent(lua)

		val metadata = UniqueEffectNetworkedValues()
		val metadataNetworkID = uniqueEffectMetadata.add(metadata)

		var isPersistent = duration == null
			private set

		init {
			metadata.duration = duration ?: 0.0
			metadata.sourceEntity = source
		}

		val animator: EffectAnimator?
		val animatorNetworkID: Int?

		init {
			if (effect.value.animationConfig?.fullPath != null) {
				animator = EffectAnimator(KOptional(effect.value.animationConfig!!.fullPath!!))
				animatorNetworkID = effectAnimators.add(animator)
			} else {
				animator = null
				animatorNetworkID = null
			}

			if (!entity.isRemote && effect.value.scripts.isNotEmpty()) {
				// unique effects shouldn't be initialized on remotes in first place
				// but whatever
				lua.attach(effect.value.scripts)

				// provideStatusControllerBindings(this@StatusController, lua) // provided through provideEntityBindings
				provideEntityBindings(entity, lua)
				luaMovement.init(lua)
				provideEffectBindings()

				if (animator != null) {
					// FIXME: effect animator's animator is not set in stone because of legacy protocol
					//  god damn it
					//  But at least it shouldn't change in this context
					provideAnimatorBindings(animator.animator, lua)
				}

				provideConfigBindings(lua) { path, default ->
					path.find(effect.json) ?: default
				}

				lua.init()
			}
		}

		private fun provideEffectBindings() {
			val callbacks = lua.newTable()
			lua.globals["effect"] = callbacks

			callbacks["duration"] = luaFunction {
				returnBuffer.setTo(metadata.duration)
			}

			callbacks["modifyDuration"] = luaFunction { duration: Number ->
				val value = duration.toDouble()
				check(value.isFinite()) { "Infinite duration provided" }
				check(!value.isNaN()) { "NaN duration provided" }

				if (!isPersistent) {
					metadata.duration += value
				}
			}

			callbacks["expire"] = luaFunction {
				if (!isPersistent) {
					metadata.duration = 0.0
				}
			}

			callbacks["sourceEntity"] = luaFunction {
				returnBuffer.setTo(metadata.sourceEntity ?: entity.entityID)
			}

			callbacks["setParentDirectives"] = luaFunction { directives: ByteString ->
				parentDirectives = directives.decode().sbIntern()
			}

			callbacks["getParameter"] = createConfigBinding { path, default ->
				path.find(effect.json) ?: default
			}

			callbacks["addStatModifierGroup"] = luaFunction { modifiers: Table ->
				val actual = modifiers
					.iterator()
					.map { (_, v) -> Starbound.gson.fromJsonFast(toJsonFromLua(v), StatModifier::class.java) }
					.collect(Collectors.toCollection(::ArrayList))

				val id = statModifiers.add(actual)
				modifierGroups.add(id)
				returnBuffer.setTo(id)
			}

			callbacks["setStatModifierGroup"] = luaFunction { groupID: Number, modifiers: Table ->
				val id = groupID.toInt()

				if (id !in modifierGroups)
					throw LuaRuntimeException("$groupID stat modifier group does not belong to this effect")

				val actual = modifiers
					.iterator()
					.map { (_, v) -> Starbound.gson.fromJsonFast(toJsonFromLua(v), StatModifier::class.java) }
					.collect(Collectors.toCollection(::ArrayList))

				statModifiers[id] = actual
			}

			callbacks["removeStatModifierGroup"] = luaFunction { groupID: Number ->
				val id = groupID.toInt()

				if (modifierGroups.rem(id)) {
					statModifiers.remove(id)
				} else {
					throw LuaRuntimeException("$groupID stat modifier group does not belong to this effect")
				}
			}
		}

		fun promoteToPersistent() {
			// It is important to note here that if a unique effect exists, it *may*
			// not come from a persistent effect, it *may* be from an ephemeral effect.
			// Here, when a persistent effect overrides an ephemeral effect, it is
			// clearing the duration making it into a solely persistent effect.  This
			// means that by applying a persistent effect and then clearing it, you can
			// remove an ephemeral effect.

			// Original engine only updates the duration (setting it to null),
			// but for proper promotion we also need to reset cause entity, otherwise
			// persistent effect will be sourced to whoever put ephemeral effect
			metadata.duration = 0.0
			metadata.sourceEntity = null
			isPersistent = true
		}

		fun remove() {
			lua.invokeGlobal("onExpire")
			uniqueEffectMetadata.remove(metadataNetworkID)

			for (group in modifierGroups) {
				statModifiersNetworkMap.remove(group)
			}

			if (animatorNetworkID != null) {
				effectAnimators.remove(animatorNetworkID)
			}

			check(uniqueStatusEffects.remove(effect.key) === this)
		}
	}

	fun addEphemeralEffect(effect: EphemeralStatusEffect, source: Int? = null) {
		val entry = effect.effect.entry ?: return LOGGER.warn("Tried to add invalid ephemeral effect $effect to status controller")
		val existing = uniqueStatusEffects[entry.key]

		if (existing == null) {
			addUniqueEffect(entry, effect.duration, source)
		} else {
			// If the effect exists and does not have a null duration, then refresh
			// the duration to the max
			if (!existing.isPersistent) {
				val newDuration = effect.duration ?: entry.value.defaultDuration

				if (newDuration > existing.metadata.duration) {
					// Only overwrite the sourceEntityId if the duration is *extended*
					existing.metadata.sourceEntity = source
					existing.metadata.duration = newDuration
				}

				existing.metadata.maxDuration = existing.metadata.maxDuration.coerceAtLeast(newDuration)
			}
		}
	}

	fun removeEphemeralEffect(effect: String): Boolean {
		val get = uniqueStatusEffects[effect]

		if (get != null && !get.isPersistent) {
			get.remove()
			return true
		}

		return false
	}

	fun removeEphemeralEffect(effect: Registry.Entry<StatusEffectDefinition>): Boolean {
		return removeEphemeralEffect(effect.key)
	}

	fun removeEphemeralEffects() {
		uniqueStatusEffects.values.filter { !it.isPersistent }.forEach { it.remove() }
	}

	private fun addUniqueEffect(effect: Registry.Entry<StatusEffectDefinition>, duration: Double?, source: Int?) {
		if (effect.key in uniqueStatusEffects) {
			throw IllegalStateException("Already has unique effect $effect")
		}

		if (
			(duration == null || isStatusEffectsImmune) && // ephemeral status effect
			(effect.value.blockingStat == null || !statPositive(effect.value.blockingStat!!)) // effect can't be combined with
		) {
			UniqueEffectInstance(effect, duration, source)
		}
	}

	fun activeUniqueStatusEffectSummary(): List<Pair<Registry.Entry<StatusEffectDefinition>, Double>> {
		val result = ArrayList<Pair<Registry.Entry<StatusEffectDefinition>, Double>>()

		for (data in uniqueStatusEffects.values) {
			if (data.isPersistent) {
				result.add(data.effect to 1.0)
			} else {
				result.add(data.effect to (data.metadata.duration / data.metadata.maxDuration).let {
					if (it.isInfinite() || it.isNaN())
						1.0
					else
						it
				})
			}
		}

		return result
	}

	fun uniqueStatusEffectActive(name: String): Boolean {
		return name in uniqueStatusEffects
	}

	private class EffectAnimator(var config: KOptional<String> = KOptional()) : Passthrough() {
		var animator = Animator()
			private set

		override val parentElement: NetworkedElement
			get() = animator.networkGroup

		override fun readInitial(data: DataInputStream, isLegacy: Boolean) {
			val old = config
			config = data.readKOptional { readInternedString() }

			if (old != config)
				animator = config.map { Animator.load(it) }.orElse { Animator() }

			super.readInitial(data, isLegacy)
		}

		override fun writeInitial(data: DataOutputStream, isLegacy: Boolean) {
			data.writeKOptional(config) { writeBinaryString(it) }
			super.writeInitial(data, isLegacy)
		}
	}

	private class UniqueEffectNetworkedValues {
		val networkGroup = NetworkedGroup()

		var duration by networkedFixedPoint(0.01).also { networkGroup.add(it); it.interpolator = Interpolator.Linear }
		var maxDuration by networkedFloat().also { networkGroup.add(it) }
		var sourceEntity by networkedData(null, IntValueCodec.nullable()).also { networkGroup.add(it) }
	}

	// ----- Stats
	// The heart of actor's information, such as health, energy, player's hunger, etc.
	// Stats are ephemeral, meaning their values are updated frequently, and max values are calculated
	// based on other values, resources, or StatModifiers
	sealed class LiveStat {
		abstract val baseValue: Double

		// Value with just the base percent modifiers applied and the value
		// modifiers
		abstract val baseModifiedValue: Double

		// Final modified value that includes the effective modifiers.
		abstract val effectiveModifiedValue: Double
	}

	private class LiveStatImpl : LiveStat() {
		override var baseValue: Double = 0.0

		var tickVisited = -1

		// Value with just the base percent modifiers applied and the value
		// modifiers
		override var baseModifiedValue: Double = 0.0

		// Final modified value that includes the effective modifiers.
		override var effectiveModifiedValue: Double = 0.0
	}

	sealed class Resource {
		// null means no limit
		abstract val max: Either<String, Double>?
		abstract val delta: Either<String, Double>?

		abstract val name: String

		abstract var isLocked: Boolean
		abstract var value: Double

		abstract val maxValue: Double?

		abstract fun reset()

		val isPositive: Boolean
			get() = value > 0.0

		val percentage: Double?
			get() {
				val max = maxValue ?: return null
				if (max == 0.0) return null
				return (value / max).coerceIn(0.0, 1.0)
			}

		fun give(amount: Double): Double {
			val old = value
			value += amount
			return value - old
		}

		fun consume(amount: Double, allowOverdraw: Boolean = false): Boolean {
			require(amount >= 0.0) { "Tried to consume negative amount $amount of resource $name" }

			if (isLocked) {
				return false
			} else if (value >= amount) {
				value -= amount
				return true
			} else if (value > 0.0 && allowOverdraw) {
				value = 0.0
				return true
			} else {
				return false
			}
		}

		fun setAsPercentage(percent: Double) {
			val maxValue = maxValue

			if (maxValue == null) {
				if (percent == 0.0) {
					LOGGER.warn("Preserving odd quirk of original engine: $name requires resource $max to be present to be set as 0% of that, however, $max is missing. Due to flawed `if` branch in original engine, `initialPercentage` specified as `0` falls through condition and is treated as `initialValue`=`0`")
				} else {
					throw IllegalArgumentException("$name does not have max value")
				}
			} else {
				this.value = maxValue * percent
			}
		}

		fun modifyPercentage(delta: Double) {
			val max = maxValue  ?: throw IllegalArgumentException("$name does not have max value")

			if (max == 0.0) {
				LOGGER.debug("Tried to divide by zero resource {} because its max value is zero", name)
				return
			}

			setAsPercentage(value / max + delta)
		}
	}

	private class ResourceImpl(override val name: String, override val max: Either<String, Double>?, override val delta: Either<String, Double>?) : Resource() {
		val actualValue = networkedFloat()

		override var value: Double
			get() = actualValue.value
			set(value) {
				val maxValue = maxValue

				if (maxValue == null) {
					actualValue.value = value
				} else {
					actualValue.value = value.coerceAtMost(maxValue)
				}
			}

		val actualIsLocked = networkedBoolean()

		override var isLocked: Boolean
			get() = actualIsLocked.value
			set(value) { actualIsLocked.value = value }

		override var maxValue: Double? = null
		private var defaultValue: Double = 0.0

		fun mark() {
			defaultValue = value
		}

		override fun reset() {
			value = defaultValue
		}
	}

	// in original code it is named effectiveStats
	private val liveStatsInternal = HashMap<String, LiveStatImpl>()

	private val statModifiersNetworkMap = NetworkedMap(
		IntValueCodec,
		NATIVE_MODIFIERS_CODEC to LEGACY_MODIFIERS_CODEC,
		map = ListenableMap(Int2ObjectAVLTreeMap()))

	private val statModifiers = IdMap(map = statModifiersNetworkMap)
	private val resourcesInternal = HashMap<String, ResourceImpl>()

	val liveStats: Map<String, LiveStat> = Collections.unmodifiableMap(liveStatsInternal)
	val resources: Map<String, Resource> = Collections.unmodifiableMap(resourcesInternal)

	init {
		for ((statName, stat) in config.stats) {
			val live = LiveStatImpl()
			liveStatsInternal[statName] = live
			live.baseValue = stat.baseValue
			live.baseModifiedValue = stat.baseValue
		}

		for ((resourceName, res) in config.resources) {
			resourcesInternal[resourceName] = ResourceImpl(
				resourceName,
				max = res.maxValue?.let { Either.right(it) } ?: res.maxStat?.let { Either.left(it) },
				delta = res.deltaValue?.let { Either.right(it) } ?: res.deltaStat?.let { Either.left(it) },
			)
		}

		updateStats(0.0)

		for ((resourceName, res) in config.resources) {
			val resource = resourcesInternal[resourceName]!!

			if (res.initialValue != null) {
				resource.value = res.initialValue
			} else if (res.initialPercentage != null) {
				resource.setAsPercentage(res.initialPercentage)
			} else if (resource.maxValue != null) {
				resource.setAsPercentage(1.0)
			}

			resource.mark()
		}

		statNetworkGroup.add(statModifiersNetworkMap)

		for (k in resourcesInternal.keys.sorted()) {
			val resource = resourcesInternal[k]!!

			statNetworkGroup.add(resource.actualValue)
			statNetworkGroup.add(resource.actualIsLocked)
		}
	}

	fun resetResources() {
		for (resource in resourcesInternal.values) {
			resource.reset()
		}
	}

	fun statPositive(name: String): Boolean {
		return (liveStatsInternal[name]?.effectiveModifiedValue ?: 0.0) > 0.0
	}

	private var statsTick = 0

	private fun updateStats(delta: Double) {
		// We use two intermediate values for calculating the effective stat value.
		// The baseModifiedValue represents the application of the base percentage
		// modifiers and the value modifiers, which only depend on the baseValue.
		// The effectiveModifiedValue is the application of all effective percentage
		// modifiers successively on the baseModifiedValue, causing them to stack with
		// each other in addition to base multipliers and value modifiers

		statsTick++

		for ((statName, stat) in config.stats) {
			val live = liveStatsInternal[statName]!!
			live.baseValue = stat.baseValue
			live.baseModifiedValue = stat.baseValue
			live.tickVisited = statsTick
		}

		for (group in statModifiers.values) {
			for (modifier in group) {
				var live = liveStatsInternal[modifier.stat]

				if (live == null) {
					live = LiveStatImpl()
					liveStatsInternal[modifier.stat] = live
				}

				live.tickVisited = statsTick

				if (modifier.type == StatModifierType.BASE_MULTIPLICATION) {
					live.baseModifiedValue += (modifier.value - 1.0) * live.baseValue
				} else if (modifier.type == StatModifierType.BASE_ADDITION) {
					live.baseModifiedValue += modifier.value
				}
			}
		}

		// Then we do all the StatEffectiveMultipliers and compute the
		// final effectiveModifiedValue

		for (value in liveStatsInternal.values)
			value.effectiveModifiedValue = value.baseModifiedValue

		for (group in statModifiers.values) {
			for (modifier in group) {
				val live = liveStatsInternal[modifier.stat]!!

				if (modifier.type == StatModifierType.OVERALL_MULTIPLICATION) {
					live.effectiveModifiedValue *= modifier.value
				} else if (modifier.type == StatModifierType.OVERALL_ADDITION) {
					live.effectiveModifiedValue += modifier.value
				}
			}
		}

		for (live in liveStatsInternal.values) {
			if (live.tickVisited != statsTick) {
				live.baseValue = 0.0
				live.effectiveModifiedValue = 0.0
				live.baseModifiedValue = 0.0
			}
		}

		// Then update all the resources due to charging and percentage tracking,
		// after updating the stats.

		for (resource in resourcesInternal.values) {
			val oldMax = resource.maxValue
			resource.maxValue = resource.max?.map({ liveStatsInternal[it]?.effectiveModifiedValue }, { it })

			// If the resource has a maximum value, rather than keeping the absolute
			// value of the resource the same between updates, the resource value
			// should instead track the percentage.
			if (oldMax != null && resource.maxValue != null && resource.maxValue!! > 0.0) {
				resource.value *= resource.maxValue!! / oldMax
			}

			if (resource.maxValue != null) {
				resource.value = resource.value
			}

			if (delta > 0.0) {
				resource.value += delta * (resource.delta?.map({ liveStatsInternal[it]?.effectiveModifiedValue }, { it }) ?: 0.0)
			}
		}
	}

	fun tick(delta: Double) {
		updateStats(delta)
		animator?.animator?.tick(delta, entity.world.random)

		recentHitsDealt.tick(delta)
		recentDamageDealt.tick(delta)
		recentDamageReceived.tick(delta)

		luaUpdate.update(delta) {
			luaMovement.clearControlsIfNeeded()
		}

		val isStatusEffectsImmune = isStatusEffectsImmune

		if (!isStatusEffectsImmune && entity.movement.liquidPercentage > minimumLiquidStatusEffectPercentage) {
			entity.movement.liquid?.value?.statusEffects?.forEach {
				val entry = it.entry

				if (entry != null)
					addEphemeralEffect(EphemeralStatusEffect(entry, entry.value.defaultDuration))
			}
		}

		if (environmentalUpdateTimer.wrapTick(delta)) {
			val entityEffects = ArrayList<PersistentStatusEffect>()

			if (!isStatusEffectsImmune) {
				val entities = ReferenceOpenHashSet<StatusEffectEntity>()

				for (poly in entity.movement.computeGlobalHitboxes()) {
					entity.world.entityIndex.iterate(poly.aabb, {
						if (
							it is StatusEffectEntity &&
							entities.add(it) &&
							it.statusEffectArea.any { p -> entity.world.geometry.polyIntersectsPoly(p, poly) }
						) {
							entityEffects.addAll(it.entityEffects)
						}
					})
				}
			}

			setPersistentEffects("entities", entityEffects)

			if (!isStatusEffectsImmune && appliesEnvironmentStatusEffects) {
				setPersistentEffects("environment", entity.world.environmentStatusEffects(entity.movement.position))
			} else {
				setPersistentEffects("environment", listOf())
			}

			if (!isStatusEffectsImmune && appliesWeatherStatusEffects) {
				entity.world.weatherStatusEffects(entity.movement.position).forEach {
					addEphemeralEffect(it)
				}
			}
		}

		val toRemove = ArrayList<UniqueEffectInstance>()

		for (effect in uniqueStatusEffects.values) {
			effect.luaUpdate.update(delta) {
				luaMovement.clearControlsIfNeeded()
			}

			if (!effect.isPersistent) {
				effect.metadata.duration -= delta
			}

			if (
				!effect.isPersistent && effect.metadata.duration <= 0.0 ||
				!effect.isPersistent && isStatusEffectsImmune ||
				effect.effect.value.blockingStat != null && statPositive(effect.effect.value.blockingStat!!)
			) {
				toRemove.add(effect)
			}
		}

		toRemove.forEach {
			it.remove()
		}

		val parentDirectives = StringBuilder(primaryDirectives)

		for (effect in uniqueStatusEffects.values) {
			parentDirectives.append("?")
			parentDirectives.append(effect.parentDirectives)
		}

		this.parentDirectives = parentDirectives.toString()
	}

	fun tickRemote(delta: Double) {

	}

	companion object {
		private val LEGACY_MODIFIERS_CODEC = StreamCodec.Collection(StatModifier.LEGACY_CODEC, ::ArrayList)
		private val NATIVE_MODIFIERS_CODEC = StreamCodec.Collection(StatModifier.CODEC, ::ArrayList)
		private val LOGGER = LogManager.getLogger()
	}
}
